@@ -31,9 +31,11 @@
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif
+#include <stdio.h>
 
 #include <mysql.h>
 #include <mysqld_error.h>
+#include <errmsg.h>
 
 #include "sb_options.h"
 #include "db_driver.h"
@@ -50,7 +52,12 @@
 # define ENGINE_CLAUSE "TYPE"
 #endif
 
-#define DEBUG(format, ...) do { if (db_globals.debug) log_text(LOG_DEBUG, format, __VA_ARGS__); } while (0)
+#define DEBUG(format, ...) do { if (args.debug) log_text(LOG_DEBUG, format, __VA_ARGS__); } while (0)
+
+#define SAFESTR(s) ((s != NULL) ? (s) : "(null)")
+
+/* FIXME */
+db_bind_t *gresults;
 
 /* MySQL driver arguments */
 
@@ -58,7 +65,7 @@ static sb_arg_t mysql_drv_args[] =
 {
   {"mysql-host", "MySQL server host", SB_ARG_TYPE_LIST, "localhost"},
   {"mysql-port", "MySQL server port", SB_ARG_TYPE_INT, "3306"},
-  {"mysql-socket", "MySQL socket", SB_ARG_TYPE_STRING, NULL},
+  {"mysql-socket", "MySQL socket", SB_ARG_TYPE_LIST, NULL},
   {"mysql-user", "MySQL user", SB_ARG_TYPE_STRING, "sbtest"},
   {"mysql-password", "MySQL password", SB_ARG_TYPE_STRING, ""},
   {"mysql-db", "MySQL database name", SB_ARG_TYPE_STRING, "sbtest"},
@@ -68,32 +75,37 @@ static sb_arg_t mysql_drv_args[] =
    SB_ARG_TYPE_STRING, "auto"},
   {"mysql-ssl", "use SSL connections, if available in the client library", SB_ARG_TYPE_FLAG, "off"},
   {"myisam-max-rows", "max-rows parameter for MyISAM tables", SB_ARG_TYPE_INT, "1000000"},
-  {"mysql-create-options", "additional options passed to CREATE TABLE", SB_ARG_TYPE_STRING, ""},
-  
+  {"mysql-debug", "dump all client library calls", SB_ARG_TYPE_FLAG, "off"},
+  {"mysql-ignore-errors", "list of errors to ignore, or \"all\"",
+   SB_ARG_TYPE_LIST, "1213,1020,1205"},
+
   {NULL, NULL, SB_ARG_TYPE_NULL, NULL}
 };
 
-typedef enum
-{
-  ENGINE_TRX_YES,
-  ENGINE_TRX_NO,
-  ENGINE_TRX_AUTO
-} mysql_drv_trx_t;
-
 typedef struct
 {
-  sb_list_t         *hosts;
+  sb_list_t          *hosts;
   unsigned int       port;
-  char               *socket;
+  sb_list_t          *sockets;
   char               *user;
   char               *password;
   char               *db;
-  unsigned int       myisam_max_rows;
-  mysql_drv_trx_t    engine_trx;
-  unsigned int       use_ssl;
-  char               *create_options;
+  unsigned char      use_ssl;
+  unsigned char      debug;
+  sb_list_t          *ignored_errors;
 } mysql_drv_args_t;
 
+typedef struct
+{
+  MYSQL        *mysql;
+  char         *host;
+  char         *user;
+  char         *password;
+  char         *db;
+  unsigned int port;
+  char         *socket;
+} db_mysql_conn_t;
+
 #ifdef HAVE_PS
 /* Structure used for DB-to-MySQL bind types map */
 
@@ -127,31 +139,28 @@ static drv_caps_t mysql_drv_caps =
 {
   1,
   0,
-  0,
   1,
   0,
   0,
-  1,
-  
-  NULL
+  1
 };
 
 
 
 static mysql_drv_args_t args;          /* driver args */
 
-static char table_options_str[1024];   /* table options string */
-
 static char use_ps; /* whether server-side prepared statemens should be used */
 
+/* Positions in the list of hosts/sockets protected by hosts_mutex */
 static sb_list_item_t *hosts_pos;
+static sb_list_item_t *sockets_pos;
 
 static pthread_mutex_t hosts_mutex;
 
 /* MySQL driver operations */
 
 static int mysql_drv_init(void);
-static int mysql_drv_describe(drv_caps_t *, const char *);
+static int mysql_drv_describe(drv_caps_t *);
 static int mysql_drv_connect(db_conn_t *);
 static int mysql_drv_disconnect(db_conn_t *);
 static int mysql_drv_prepare(db_stmt_t *, const char *);
@@ -192,13 +201,12 @@ static db_driver_t mysql_driver =
     mysql_drv_store_results,
     mysql_drv_done
   },
-  {NULL, NULL}
+  {0,0}
 };
 
 
 /* Local functions */
 
-static int parse_table_engine(const char *);
 #ifdef HAVE_PS
 static int get_mysql_bind_type(db_bind_type_t);
 #endif
@@ -219,8 +227,6 @@ int register_driver_mysql(sb_list_t *drivers)
 
 int mysql_drv_init(void)
 {
-  char *s;
-  
   args.hosts = sb_get_value_list("mysql-host");
   if (SB_LIST_IS_EMPTY(args.hosts))
   {
@@ -229,18 +235,21 @@ int mysql_drv_init(void)
   }
   hosts_pos = args.hosts;
   pthread_mutex_init(&hosts_mutex, NULL);
-  
-  args.port = sb_get_value_int("mysql-port");
-  args.socket = sb_get_value_string("mysql-socket");
+
+  args.sockets = sb_get_value_list("mysql-socket");
+  sockets_pos = args.sockets;
+
+  args.port = (unsigned int)sb_get_value_int("mysql-port");
   args.user = sb_get_value_string("mysql-user");
   args.password = sb_get_value_string("mysql-password");
   args.db = sb_get_value_string("mysql-db");
-  args.myisam_max_rows = sb_get_value_int("myisam-max-rows");
   args.use_ssl = sb_get_value_flag("mysql-ssl");
-  args.create_options = sb_get_value_string("mysql-create-options");
-  if (args.create_options == NULL)
-    args.create_options = "";
-  
+  args.debug = sb_get_value_flag("mysql-debug");
+  if (args.debug)
+    sb_globals.verbosity = LOG_DEBUG;
+
+  args.ignored_errors = sb_get_value_list("mysql-ignore-errors");
+
   use_ps = 0;
 #ifdef HAVE_PS
   mysql_drv_caps.prepared_statements = 1;
@@ -248,145 +257,31 @@ int mysql_drv_init(void)
     use_ps = 1;
 #endif
 
-  s = sb_get_value_string("mysql-engine-trx");
-  if (s == NULL)
-  {
-    log_text(LOG_FATAL, "--mysql-engine-trx requires an argument");
-    return 1;
-  }
-  if (!strcasecmp(s, "yes"))
-    args.engine_trx = ENGINE_TRX_YES;
-  else if (!strcasecmp(s, "no"))
-    args.engine_trx = ENGINE_TRX_NO;
-  else if (!strcasecmp(s, "auto"))
-    args.engine_trx = ENGINE_TRX_AUTO;
-  else
-  {
-    log_text(LOG_FATAL, "Invalid value for mysql-engine-trx: %s", s);
-    return 1;
-  }
-  
-  s = sb_get_value_string("mysql-table-engine");
-
+  DEBUG("mysql_library_init(%d, %p, %p)", 0, NULL, NULL);
   mysql_library_init(0, NULL, NULL);
   
-  return parse_table_engine(s);
+  return 0;
 }
 
 
-/* Describe database capabilities (possibly depending on table type) */
+/* Describe database capabilities */
 
 
-int mysql_drv_describe(drv_caps_t *caps, const char * table_name)
+int mysql_drv_describe(drv_caps_t *caps)
 {
-  db_conn_t    con;
-  char         query[1024];
-  MYSQL_RES    *res = NULL;
-  MYSQL_ROW    row;
-  MYSQL_FIELD  *fields;
-  unsigned int num_fields;
-  unsigned int i;
-  int          connected = 0;
-  int          rc = 0;
-  
   *caps = mysql_drv_caps;
   
-  if (table_name == NULL)
-    goto exit;
-
-  /* Try to determine table type */
-  if (mysql_drv_connect(&con))
-    goto error;
-
-  /* Fix the hosts list */
-  hosts_pos = args.hosts;
-  
-  connected = 1;
-  snprintf(query, sizeof(query), "SHOW TABLE STATUS LIKE '%s'", table_name);
-  
-  rc = mysql_real_query(con.ptr, query, strlen(query)); 
-  DEBUG("mysql_real_query(%p, \"%s\", %d) = %d", con.ptr, query, strlen(query), rc);
-  if (rc)
-    goto error;
-  
-  res = mysql_store_result(con.ptr);
-  DEBUG("mysql_store_result(%p) = %p", con.ptr, res);
-  if (res == NULL)
-    goto error;
-
-  num_fields = mysql_num_fields(res);
-  DEBUG("mysql_num_fields(%p) = %d", res, num_fields);
-  fields = mysql_fetch_fields(res);
-  DEBUG("mysql_fetch_fields(%p) = %p", res, fields);
-  for (i = 0; i < num_fields; i++)
-    if (!strcasecmp(fields[i].name, "type") ||
-        !strcasecmp(fields[i].name, "engine"))
-      break;
-  if (i >= num_fields)
-    goto error;
-  
-  row = mysql_fetch_row(res);
-  DEBUG("mysql_fetch_row(%p) = %p", res, row);
-  if (row == NULL || row[i] == NULL)
-    goto error;
-  
-  rc = parse_table_engine(row[i]);
-  *caps = mysql_drv_caps;
-  goto exit;
-
- error:
-  log_text(LOG_ALERT, "Error: failed to determine table '%s' type!", table_name);
-  log_text(LOG_ALERT, "MySQL error: %s", mysql_error(con.ptr));
-  rc = 1;
-  
- exit:
-  if (res != NULL)
-  {
-    mysql_free_result(res);
-    DEBUG("mysql_free_result(%p)", res);
-  }
-  if (connected)
-    mysql_drv_disconnect(&con);
-
-  return rc;
+  return 0;
 }
 
 
-/* Connect to MySQL database */
-
-
-int mysql_drv_connect(db_conn_t *sb_conn)
+static int mysql_drv_real_connect(db_mysql_conn_t *db_mysql_con)
 {
-  MYSQL          *con;
-  char           *host;
-  char           *ssl_key;
-  char           *ssl_cert;
-  char           *ssl_ca;
-  
-  con = (MYSQL *)malloc(sizeof(MYSQL));
-  if (con == NULL)
-    return 1;
-  sb_conn->ptr = con;
-  
-  mysql_init(con);
-  DEBUG("mysql_init(%p)", con);
+  MYSQL          *con = db_mysql_con->mysql;
+  const char     *ssl_key;
+  const char     *ssl_cert;
+  const char     *ssl_ca;
 
-  pthread_mutex_lock(&hosts_mutex);
-  hosts_pos = SB_LIST_ITEM_NEXT(hosts_pos);
-  if (hosts_pos == args.hosts)
-    hosts_pos = SB_LIST_ITEM_NEXT(hosts_pos);
-  host = SB_LIST_ENTRY(hosts_pos, value_t, listitem)->data;
-  pthread_mutex_unlock(&hosts_mutex);
-  
-#if 0
-  /*
-    FIXME: the following leads to crash in the client lib.
-    http://bugs.mysql.com/?id=40552
-  */
-  mysql_options(con, MYSQL_READ_DEFAULT_GROUP, "sysbench");
-  DEBUG("mysql_options(%p, MYSQL_READ_DEFAULT_GROUP, \"sysbench\")", con);
-#endif
-  
   if (args.use_ssl)
   {
     ssl_key= "client-key.pem";
@@ -397,38 +292,93 @@ int mysql_drv_connect(db_conn_t *sb_conn)
           ssl_cert, ssl_ca);
     mysql_ssl_set(con, ssl_key, ssl_cert, ssl_ca, NULL, NULL);
   }
-  
+
   DEBUG("mysql_real_connect(%p, \"%s\", \"%s\", \"%s\", \"%s\", %u, \"%s\", %s)",
         con,
-        host,
-        args.user,
-        args.password,
-        args.db,
-        args.port,
-        args.socket,
+        SAFESTR(db_mysql_con->host),
+        SAFESTR(db_mysql_con->user),
+        SAFESTR(db_mysql_con->password),
+        SAFESTR(db_mysql_con->db),
+        db_mysql_con->port,
+        SAFESTR(db_mysql_con->socket),
         (MYSQL_VERSION_ID >= 50000) ? "CLIENT_MULTI_STATEMENTS" : "0"
         );
-  if (!mysql_real_connect(con,
-                         host,
-                         args.user,
-                         args.password,
-                         args.db,
-                         args.port,
-                         args.socket,
+
+  return mysql_real_connect(con,
+                            db_mysql_con->host,
+                            db_mysql_con->user,
+                            db_mysql_con->password,
+                            db_mysql_con->db,
+                            db_mysql_con->port,
+                            db_mysql_con->socket,
 #if MYSQL_VERSION_ID >= 50000
-                          CLIENT_MULTI_STATEMENTS)
+                            CLIENT_MULTI_STATEMENTS
 #else
-                          0)
+                            0
 #endif
-      )
+                            ) == NULL;
+}
+
+
+/* Connect to MySQL database */
+
+
+int mysql_drv_connect(db_conn_t *sb_conn)
+{
+  MYSQL           *con;
+  db_mysql_conn_t *db_mysql_con;
+
+  db_mysql_con = (db_mysql_conn_t *) calloc(1, sizeof(db_mysql_conn_t));
+
+  if (db_mysql_con == NULL)
+    return 1;
+
+  con = (MYSQL *) malloc(sizeof(MYSQL));
+  if (con == NULL)
+    return 1;
+
+  db_mysql_con->mysql = con;
+
+  DEBUG("mysql_init(%p)", con);
+  mysql_init(con);
+
+  pthread_mutex_lock(&hosts_mutex);
+
+  if (SB_LIST_IS_EMPTY(args.sockets))
+  {
+    db_mysql_con->socket = NULL;
+    hosts_pos = SB_LIST_ITEM_NEXT(hosts_pos);
+    if (hosts_pos == args.hosts)
+      hosts_pos = SB_LIST_ITEM_NEXT(hosts_pos);
+    db_mysql_con->host = SB_LIST_ENTRY(hosts_pos, value_t, listitem)->data;
+  }
+  else
+  {
+    db_mysql_con->host = "localhost";
+    sockets_pos = SB_LIST_ITEM_NEXT(sockets_pos);
+    if (sockets_pos == args.sockets)
+      sockets_pos = SB_LIST_ITEM_NEXT(sockets_pos);
+    db_mysql_con->socket = SB_LIST_ENTRY(sockets_pos, value_t, listitem)->data;
+  }
+  pthread_mutex_unlock(&hosts_mutex);
+
+  db_mysql_con->user = args.user;
+  db_mysql_con->password = args.password;
+  db_mysql_con->db = args.db;
+  db_mysql_con->port = args.port;
+
+  if (mysql_drv_real_connect(db_mysql_con))
   {
     log_text(LOG_FATAL, "unable to connect to MySQL server, aborting...");
     log_text(LOG_FATAL, "error %d: %s", mysql_errno(con),
-           mysql_error(con));
+             mysql_error(con));
+    free(db_mysql_con);
     free(con);
     return 1;
   }
 
+  sb_conn->ptr = db_mysql_con;
+
   return 0;
 }
 
@@ -438,15 +388,16 @@ int mysql_drv_connect(db_conn_t *sb_conn)
 
 int mysql_drv_disconnect(db_conn_t *sb_conn)
 {
-  MYSQL *con = sb_conn->ptr;
+  db_mysql_conn_t *db_mysql_con = sb_conn->ptr;
 
-  if (con != NULL)
+  if (db_mysql_con != NULL && db_mysql_con->mysql != NULL)
   {
-    DEBUG("mysql_close(%p)", con);
-    mysql_close(con);
+    DEBUG("mysql_close(%p)", db_mysql_con->mysql);
+    mysql_close(db_mysql_con->mysql);
+    free(db_mysql_con->mysql);
+    free(db_mysql_con);
   }
-  free(con);
-  
+
   return 0;
 }
 
@@ -457,7 +408,8 @@ int mysql_drv_disconnect(db_conn_t *sb_conn)
 int mysql_drv_prepare(db_stmt_t *stmt, const char *query)
 {
 #ifdef HAVE_PS
-  MYSQL      *con = (MYSQL *)stmt->connection->ptr;
+  db_mysql_conn_t *db_mysql_con = (db_mysql_conn_t *) stmt->connection->ptr;
+  MYSQL      *con = db_mysql_con->mysql;
   MYSQL_STMT *mystmt;
   unsigned int rc;
 
@@ -482,9 +434,9 @@ int mysql_drv_prepare(db_stmt_t *stmt, const char *query)
       DEBUG("mysql_errno(%p) = %u", con, rc);
       if (rc == ER_UNSUPPORTED_PS)
       {
-        log_text(LOG_DEBUG,
-                 "Failed to prepare query \"%s\", using emulation",
-                 query);
+        log_text(LOG_INFO,
+                 "Failed to prepare query \"%s\" (%d: %s), using emulation",
+                 query, rc, mysql_error(con));
         goto emulate;
       }
       else
@@ -497,7 +449,8 @@ int mysql_drv_prepare(db_stmt_t *stmt, const char *query)
         return 1;
       }
     }
-
+    stmt->query = strdup(query);
+    
     return 0;
   }
 
@@ -517,11 +470,13 @@ int mysql_drv_prepare(db_stmt_t *stmt, const char *query)
 
 int mysql_drv_bind_param(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
 {
-  MYSQL        *con = (MYSQL *)stmt->connection->ptr;
+  db_mysql_conn_t *db_mysql_con = (db_mysql_conn_t *) stmt->connection->ptr;
+  MYSQL        *con = db_mysql_con->mysql;
 #ifdef HAVE_PS
   MYSQL_BIND   *bind;
   unsigned int i;
-  unsigned int rc;
+  my_bool rc;
+  unsigned long param_count;
 #endif
   
   if (con == NULL)
@@ -533,11 +488,13 @@ int mysql_drv_bind_param(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
     if (stmt->ptr == NULL)
       return 1;
     /* Validate parameters count */
-    rc = mysql_stmt_param_count(stmt->ptr);
-    DEBUG("mysql_stmt_param_count(%p) = %u", stmt->ptr, rc);
-    if (rc != len)
+    param_count = mysql_stmt_param_count(stmt->ptr);
+    DEBUG("mysql_stmt_param_count(%p) = %lu", stmt->ptr, param_count);
+    if (param_count != len)
+    {
+      log_text(LOG_FATAL, "Wrong number of parameters to mysql_stmt_bind_param");
       return 1;
-
+    }
     /* Convert SysBench bind structures to MySQL ones */
     bind = (MYSQL_BIND *)calloc(len, sizeof(MYSQL_BIND));
     if (bind == NULL)
@@ -552,7 +509,7 @@ int mysql_drv_bind_param(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
     }
 
     rc = mysql_stmt_bind_param(stmt->ptr, bind);
-    DEBUG("mysql_stmt_bind_param(%p, %p) = %u", stmt->ptr, bind, rc);
+    DEBUG("mysql_stmt_bind_param(%p, %p) = %d", stmt->ptr, bind, rc);
     if (rc)
     {
       log_text(LOG_FATAL, "mysql_stmt_bind_param() failed");
@@ -594,10 +551,11 @@ int mysql_drv_bind_result(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
 
   return 1;
 #else
-  MYSQL        *con = (MYSQL *)stmt->connection->ptr;
+  db_mysql_conn_t *db_mysql_con =(db_mysql_conn_t *) stmt->connection->ptr;
+  MYSQL        *con = db_mysql_con->mysql;
   MYSQL_BIND   *bind;
   unsigned int i;
-  unsigned int rc;
+  my_bool rc;
   
   if (con == NULL || stmt->ptr == NULL)
     return 1;
@@ -615,8 +573,11 @@ int mysql_drv_bind_result(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
     bind[i].is_null = params[i].is_null;
   }
 
+  /* FIXME */
+  gresults = params;
+  
   rc = mysql_stmt_bind_result(stmt->ptr, bind);
-  DEBUG("mysql_stmt_bind_result(%p, %p) = %u", stmt->ptr, bind, rc);
+  DEBUG("mysql_stmt_bind_result(%p, %p) = %d", stmt->ptr, bind, rc);
   if (rc)
   {
     free(bind);
@@ -629,6 +590,95 @@ int mysql_drv_bind_result(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
 }
 
 
+/*
+  Reset connection to the server by reconnecting with the same parameters.
+*/
+
+
+static int mysql_drv_reconnect(db_conn_t *sb_con)
+{
+  db_mysql_conn_t *db_mysql_con = (db_mysql_conn_t *) sb_con->ptr;
+  MYSQL *con = db_mysql_con->mysql;
+
+  log_text(LOG_DEBUG, "Reconnecting");
+
+  DEBUG("mysql_close(%p)", con);
+  mysql_close(con);
+
+  while (mysql_drv_real_connect(db_mysql_con))
+  {
+    if (sb_globals.error)
+      return SB_DB_ERROR_FAILED;
+
+    usleep(1000);
+  }
+
+  log_text(LOG_DEBUG, "Reconnected");
+
+  return SB_DB_ERROR_RECONNECTED;
+}
+
+
+/*
+  Check if the error in a given connection should be fatal or ignored according
+  to the list of errors in --mysql-ignore-errors.
+*/
+
+
+static int check_error(db_conn_t *sb_con, const char *func, const char *query)
+{
+  sb_list_item_t *pos;
+  unsigned int   tmp;
+  unsigned int   error;
+  db_mysql_conn_t *db_mysql_con = (db_mysql_conn_t *) sb_con->ptr;
+  MYSQL          *con = db_mysql_con->mysql;
+
+  error = mysql_errno(con);
+  DEBUG("mysql_errno(%p) = %u", con, error);
+
+  /*
+    Check if the error code is specified in --mysql-ignore-errors, and return
+    SB_DB_ERROR_RESTART_TRANSACTION if so, or SB_DB_ERROR_FAILED otherwise
+  */
+  SB_LIST_FOR_EACH(pos, args.ignored_errors)
+  {
+    const char *val = SB_LIST_ENTRY(pos, value_t, listitem)->data;
+
+    tmp = (unsigned int) atoi(val);
+
+    if (error == tmp || !strcmp(val, "all"))
+    {
+      log_text(LOG_DEBUG, "Ignoring error %u %s, ", error, mysql_error(con));
+
+      /* Check if we should reconnect */
+      switch (error)
+      {
+      case CR_SERVER_LOST:
+      case CR_SERVER_GONE_ERROR:
+      case CR_TCP_CONNECTION:
+      case CR_SERVER_LOST_EXTENDED:
+
+        return mysql_drv_reconnect(sb_con);
+
+      default:
+
+        break;
+      }
+
+      return SB_DB_ERROR_RESTART_TRANSACTION;
+    }
+  }
+
+  if (query)
+    log_text(LOG_ALERT, "%s for query '%s' failed: %u %s",
+             func, query, error, mysql_error(con));
+  else
+    log_text(LOG_ALERT, "%s failed: %u %s",
+             func, error, mysql_error(con));
+
+  return SB_DB_ERROR_FAILED;
+}
+
 /* Execute prepared statement */
 
 
@@ -641,7 +691,7 @@ int mysql_drv_execute(db_stmt_t *stmt, db_result_set_t *rs)
   char            need_realloc;
   int             n;
   unsigned int    rc;
- 
+
 #ifdef HAVE_PS
   (void)rs; /* unused */
 
@@ -652,20 +702,12 @@ int mysql_drv_execute(db_stmt_t *stmt, db_result_set_t *rs)
       log_text(LOG_DEBUG, "ERROR: exiting mysql_drv_execute(), uninitialized statement");
       return SB_DB_ERROR_FAILED;
     }
-    rc = mysql_stmt_execute(stmt->ptr);
+    rc = (unsigned int)mysql_stmt_execute(stmt->ptr);
     DEBUG("mysql_stmt_execute(%p) = %u", stmt->ptr, rc);
+
     if (rc)
-    {
-      rc = mysql_errno(con->ptr);
-      DEBUG("mysql_errno(%p) = %u", con->ptr, rc);
-      if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT || 
-          rc == ER_CHECKREAD)
-        return SB_DB_ERROR_DEADLOCK;
-      log_text(LOG_ALERT, "failed to execute mysql_stmt_execute(): Err%d %s",
-               mysql_errno(con->ptr),
-               mysql_error(con->ptr));
-      return SB_DB_ERROR_FAILED;
-    }
+      return check_error(con, "mysql_stmt_execute()", stmt->query);
+
     return SB_DB_ERROR_NONE;
   }
 #else
@@ -697,13 +739,13 @@ int mysql_drv_execute(db_stmt_t *stmt, db_result_set_t *rs)
       continue;
     }
 
-    n = db_print_value(stmt->bound_param + vcnt, buf + j, buflen - j);
+    n = db_print_value(stmt->bound_param + vcnt, buf + j, (int)(buflen - j));
     if (n < 0)
     {
       need_realloc = 1;
       goto again;
     }
-    j += n;
+    j += (unsigned int)n;
     vcnt++;
   }
   buf[j] = '\0';
@@ -726,22 +768,16 @@ int mysql_drv_execute(db_stmt_t *stmt, db_result_set_t *rs)
 int mysql_drv_query(db_conn_t *sb_conn, const char *query,
                       db_result_set_t *rs)
 {
-  MYSQL *con = sb_conn->ptr;
+  db_mysql_conn_t *db_mysql_con = (db_mysql_conn_t *)sb_conn->ptr;
+  MYSQL *con = db_mysql_con->mysql;
   unsigned int rc;
 
   (void)rs; /* unused */
-  rc = mysql_real_query(con, query, strlen(query));
+  rc = (unsigned int)mysql_real_query(con, query, strlen(query));
   DEBUG("mysql_real_query(%p, \"%s\", %u) = %u", con, query, strlen(query), rc);
+
   if (rc)
-  {
-    rc = mysql_errno(con);
-    DEBUG("mysql_errno(%p) = %u", con, rc);
-    if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT || rc == ER_CHECKREAD)
-      return SB_DB_ERROR_DEADLOCK;
-    log_text(LOG_ALERT, "failed to execute MySQL query: `%s`:", query);
-    log_text(LOG_ALERT, "Error %d %s", mysql_errno(con), mysql_error(con));
-    return SB_DB_ERROR_FAILED; 
-  }  
+    return check_error(sb_conn, "mysql_drv_query()", query);
 
   return SB_DB_ERROR_NONE;
 }
@@ -753,8 +789,8 @@ int mysql_drv_query(db_conn_t *sb_conn, const char *query,
 int mysql_drv_fetch(db_result_set_t *rs)
 {
   /* NYI */
-  (void)rs;
-
+  (void)rs;  /* unused */
+  
   return 1;
 }
 
@@ -764,11 +800,17 @@ int mysql_drv_fetch(db_result_set_t *rs)
 
 int mysql_drv_fetch_row(db_result_set_t *rs, db_row_t *row)
 {
-  /* NYI */
-  (void)rs;  /* unused */
-  (void)row; /* unused */
+  db_mysql_conn_t *db_mysql_con = (db_mysql_conn_t *) rs->connection->ptr;
+  row->ptr = mysql_fetch_row(rs->ptr);
+  DEBUG("mysql_fetch_row(%p) = %p", rs->ptr, row->ptr);
+  if (row->ptr == NULL)
+  {
+    log_text(LOG_FATAL, "mysql_fetch_row() failed: %s",
+             mysql_error(db_mysql_con->mysql));
+    return 1;
+  }
   
-  return 1;
+  return 0;
 }
 
 
@@ -786,10 +828,11 @@ unsigned long long mysql_drv_num_rows(db_result_set_t *rs)
 
 int mysql_drv_store_results(db_result_set_t *rs)
 {
-  MYSQL        *con = rs->connection->ptr;
+  db_mysql_conn_t *db_mysql_con = (db_mysql_conn_t *) rs->connection->ptr;
+  MYSQL        *con = db_mysql_con->mysql;
   MYSQL_RES    *res;
   MYSQL_ROW    row;
-  int rc;
+  unsigned int rc;
 
 #ifdef HAVE_PS
   /* Is this result set from prepared statement? */
@@ -798,27 +841,19 @@ int mysql_drv_store_results(db_result_set_t *rs)
     if (rs->statement->ptr == NULL)
       return 1;
 
-    rc = mysql_stmt_store_result(rs->statement->ptr);
+    rc = (unsigned int)mysql_stmt_store_result(rs->statement->ptr);
     DEBUG("mysql_stmt_store_result(%p) = %d", rs->statement->ptr, rc);
     if (rc)
     {
-      rc = mysql_errno(con);
-      DEBUG("mysql_errno(%p) = %d", con, rc);
-      if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT)
-      {
-        log_text(LOG_WARNING,
-                 "mysql_stmt_store_result() failed with error: (%d) %s", rc,
-                 mysql_error(con));
-        return SB_DB_ERROR_DEADLOCK;
-      }
+      if (mysql_stmt_field_count(rs->statement->ptr) == 0)
+        return SB_DB_ERROR_NONE;
 
-      log_text(LOG_ALERT, "MySQL error: %s\n", mysql_error(con));
-      return SB_DB_ERROR_FAILED;
+      return check_error(rs->connection, "mysql_stmt_store_result()", NULL);
     }
     rs->nrows = mysql_stmt_num_rows(rs->statement->ptr);
     DEBUG("mysql_stmt_num_rows(%p) = %d", rs->statement->ptr, rs->nrows);
     do {
-      rc = mysql_stmt_fetch(rs->statement->ptr);
+      rc = (unsigned int)mysql_stmt_fetch(rs->statement->ptr);
       DEBUG("mysql_stmt_fetch(%p) = %d", rs->statement->ptr, rc);
     } while(rc == 0);
 
@@ -834,17 +869,10 @@ int mysql_drv_store_results(db_result_set_t *rs)
   DEBUG("mysql_store_result(%p) = %p", con, res);
   if (res == NULL)
   {
-      rc = mysql_errno(con);
-      DEBUG("mysql_errno(%p) = %u", con, rc);
-      if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT)
-      {
-        log_text(LOG_WARNING,
-                 "mysql_store_result() failed with error: (%u) %s", rc,
-                 mysql_error(con));
-        return SB_DB_ERROR_DEADLOCK;
-      }
-    log_text(LOG_ALERT, "MySQL error: %s", mysql_error(con));
-    return SB_DB_ERROR_FAILED; 
+      if (mysql_field_count(con) == 0)
+        return SB_DB_ERROR_NONE;
+
+      return check_error(rs->connection, "mysql_store_result()", NULL);
   }
   rs->ptr = (void *)res;
 
@@ -854,7 +882,6 @@ int mysql_drv_store_results(db_result_set_t *rs)
   /* just fetch result */
   while((row = mysql_fetch_row(res)))
     DEBUG("mysql_fetch_row(%p) = %p", res, row);
-
   return SB_DB_ERROR_NONE;
 }
 
@@ -889,6 +916,11 @@ int mysql_drv_free_results(db_result_set_t *rs)
 
 int mysql_drv_close(db_stmt_t *stmt)
 {
+  if (stmt->query)
+  {
+    free(stmt->query);
+    stmt->query = NULL;
+  }
 #ifdef HAVE_PS
   if (stmt->ptr == NULL)
     return 1;
@@ -911,63 +943,6 @@ int mysql_drv_done(void)
 }
 
 
-/* Parse table type and set driver capabilities */
-
-
-int parse_table_engine(const char *type)
-{
-  /* Determine if the engine supports transactions */
-  if (args.engine_trx == ENGINE_TRX_AUTO)
-  {
-    if (!strcasecmp(type, "myisam") || !strcasecmp(type, "heap") ||
-        !strcasecmp(type, "maria"))
-      mysql_drv_caps.transactions = 0;
-    else if (!strcasecmp(type, "innodb") ||
-             !strcasecmp(type, "bdb") || !strcasecmp(type, "berkeleydb") ||
-             !strcasecmp(type, "ndbcluster") ||
-             !strcasecmp(type, "federated") ||
-             !strcasecmp(type, "pbxt") ||
-             !strcasecmp(type, "falcon"))
-      mysql_drv_caps.transactions = 1;
-    else
-    {
-      log_text(LOG_FATAL, "Failed to determine transactions support for "
-               "unknown storage engine '%s'", type);
-      log_text(LOG_FATAL, "Specify explicitly with --mysql-engine-trx option");
-      return 1;
-    }
-  }
-  else
-    mysql_drv_caps.transactions = (args.engine_trx == ENGINE_TRX_YES);
-
-  /* Get engine-specific options string */
-  if (!strcasecmp(type, "myisam"))
-  {
-    if (args.myisam_max_rows > 0)
-      snprintf(table_options_str, sizeof(table_options_str),
-               "%s /*! "ENGINE_CLAUSE"=MyISAM MAX_ROWS=%u */",
-               args.create_options,
-               args.myisam_max_rows);
-    else
-      snprintf(table_options_str, sizeof(table_options_str),
-               "%s /*! "ENGINE_CLAUSE"=MyISAM */", args.create_options);
-  }
-  else if (!strcasecmp(type, "bdb") || !strcasecmp(type, "berkeleydb"))
-    snprintf(table_options_str, sizeof(table_options_str),
-             "%s /*! "ENGINE_CLAUSE"=BerkeleyDB */", args.create_options);
-  else if (!strcasecmp(type, "ndbcluster"))
-    snprintf(table_options_str, sizeof(table_options_str),
-             "%s /*! "ENGINE_CLAUSE"=NDB */", args.create_options);
-  else
-    snprintf(table_options_str, sizeof(table_options_str),
-             "%s /*! "ENGINE_CLAUSE"=%s */", args.create_options, type);
-
-  mysql_drv_caps.table_options_str = table_options_str;
-  
-  return 0;
-}
-
-
 #ifdef HAVE_PS
 /* Map SQL data type to bind_type value in MYSQL_BIND */
 
