@@ -21,6 +21,7 @@
 
 typedef struct sb_list_item_t
 {
+//    char c;
     struct sb_list_item_t *next_p;
     struct sb_list_item_t *prev_p;
 }
@@ -85,7 +86,7 @@ typedef sb_list_item_t sb_list_t ;
   ((item_p)->prev_p != (head_p))
 
 #define SB_LIST_ITEM_LAST(item_p, head_p) \
-  ((item_p)->next_p != (head_p))
+  ((item_p)->next_p == (head_p))
 
 #define SB_LIST_ITEM_NEXT(item_p) \
   ((item_p)->next_p)
@@ -94,10 +95,19 @@ typedef sb_list_item_t sb_list_t ;
   ((item_p)->prev_p)
     
 #define SB_LIST_ENTRY(ptr, type, member)            \
-((type *)((char *)(ptr) - offsetof(type, member)))
+    ((type *)(void *)(((char *)(ptr) - offsetof(type, member))))
+
+#define SB_LIST_ONCE(pos_p, head_p) \
+    pos_p= (head_p)->next_p; if (pos_p != (head_p))
 
 #define SB_LIST_FOR_EACH(pos_p, head_p) \
-    for (pos_p = (head_p)->next_p; pos_p != (head_p); pos_p = pos_p->next_p)
+  for (pos_p = (head_p)->next_p; pos_p != (head_p); pos_p = pos_p->next_p)
+
+#define SB_LIST_ENUM_START(head_p) \
+  (head_p)
+
+#define SB_LIST_ENUM_NEXT(pos_p, head_p)                         \
+  ((pos_p->next_p != (head_p)) ? (pos_p->next_p) : NULL)
 
 #define SB_LIST_FOR_EACH_SAFE(pos_p, temp_p, head_p) \
     for (pos_p = (head_p)->next_p, temp_p = (pos_p)->next_p; pos_p != (head_p); \
