@@ -29,7 +29,12 @@
 # include <ctype.h>
 #endif
 
+#ifdef HAVE_LIMITS_H
+# include <limits.h>
+#endif
+
 #include "sb_options.h"
+#include "sysbench.h"
 
 #define VALUE_DELIMITER '='
 #define VALUE_SEPARATOR ','
@@ -43,6 +48,15 @@ static sb_list_t options;
 /* List of size modifiers (kilo, mega, giga, tera) */
 static const char sizemods[] = "KMGT";
 
+/* Convert dashes to underscores in option names */
+
+static void convert_dashes(char *);
+
+/* Compare option names */
+
+static int opt_name_cmp(const char *, const char *);
+
+
 /* Initialize options library */
 
 
@@ -73,7 +87,7 @@ option_t *sb_find_option(char *name)
   return find_option(&options, name);
 }
 
-int set_option(char *name, char *value, sb_arg_type_t type)
+int set_option(const char *name, const char *value, sb_arg_type_t type)
 {
   option_t *opt;
   char     *tmp;
@@ -98,6 +112,9 @@ int set_option(char *name, char *value, sb_arg_type_t type)
       add_value(&opt->values, value);
       break;
     case SB_ARG_TYPE_LIST:
+      if (value == NULL)
+        break;
+
       tmp = strdup(value);
       for (tmp = strtok(tmp, ","); tmp != NULL; tmp = strtok(NULL, ","))
         add_value(&opt->values, tmp);
@@ -160,7 +177,13 @@ void sb_print_options(sb_arg_t *opts)
 }
 
 
-int sb_get_value_flag(char *name)
+int sb_opt_to_flag(option_t *opt)
+{
+  return !SB_LIST_IS_EMPTY(&opt->values);
+}
+
+
+int sb_get_value_flag(const char *name)
 {
   option_t *opt;
 
@@ -168,33 +191,48 @@ int sb_get_value_flag(char *name)
   if (opt == NULL)
     return 0;
 
-  return !SB_LIST_IS_EMPTY(&opt->values);
+  return sb_opt_to_flag(opt);
 }
 
 
-int sb_get_value_int(char *name)
+int sb_opt_to_int(option_t *opt)
 {
-  option_t       *opt;
   value_t        *val;
   sb_list_item_t *pos;
+  long           res;
+  char           *endptr;
+
+  SB_LIST_ONCE(pos, &opt->values)
+  {
+    val = SB_LIST_ENTRY(pos, value_t, listitem);
+    res = strtol(val->data, &endptr, 10);
+    if (*endptr != '\0' || res > INT_MAX || res < INT_MIN)
+    {
+      fprintf(stderr, "Invalid value for the '%s' option: '%s'\n",
+              opt->name, val->data);
+      exit(EXIT_FAILURE);
+    }
+    return (int) res;
+  }
+
+  return 0;
+}  
+
+
+int sb_get_value_int(const char *name)
+{
+  option_t       *opt;
 
   opt = find_option(&options, name);
   if (opt == NULL)
     return 0;
 
-  SB_LIST_FOR_EACH(pos, &opt->values)
-  {
-    val = SB_LIST_ENTRY(pos, value_t, listitem);
-    return atoi(val->data);
-  }
-
-  return 0;
+  return sb_opt_to_int(opt);;
 }
 
 
-unsigned long long sb_get_value_size(char *name)
+unsigned long long sb_opt_to_size(option_t *opt)
 {
-  option_t            *opt;
   value_t             *val;
   sb_list_item_t      *pos;
   unsigned long long  res = 0;
@@ -203,16 +241,12 @@ unsigned long long sb_get_value_size(char *name)
   unsigned int        i, n;
   char                *c;
 
-  opt = find_option(&options, name);
-  if (opt == NULL)
-    return 0;
-
-  SB_LIST_FOR_EACH(pos, &opt->values)
+  SB_LIST_ONCE(pos, &opt->values)
   {
     val = SB_LIST_ENTRY(pos, value_t, listitem);
     /*
      * Reimplentation of sscanf(val->data, "%llu%c", &res, &mult), since
-     * there is little standartization about how to specify long long values
+     * there is no standard on how to specify long long values
      */
     res = 0;
     for (rc = 0, c = val->data; *c != '\0'; c++)
@@ -230,9 +264,10 @@ unsigned long long sb_get_value_size(char *name)
       res = res * 10 + *c - '0';
     }
 
-    if (rc == 2) {
+    if (rc == 2)
+    {
       for (n = 0; sizemods[n] != '\0'; n++)
-        if (mult == sizemods[n])
+        if (toupper(mult) == sizemods[n])
           break;
       if (sizemods[n] != '\0')
       {
@@ -242,24 +277,30 @@ unsigned long long sb_get_value_size(char *name)
       else
         res = 0; /* Unknown size modifier */
     }
-    break;
   }
 
   return res;
 }
 
 
-float sb_get_value_float(char *name)
+unsigned long long sb_get_value_size(const char *name)
 {
-  option_t       *opt;
-  value_t        *val;
-  sb_list_item_t *pos;
-  float          res;
-
+  option_t            *opt;
+ 
   opt = find_option(&options, name);
   if (opt == NULL)
     return 0;
 
+  return sb_opt_to_size(opt);
+}
+
+
+float sb_opt_to_float(option_t *opt)
+{
+  value_t        *val;
+  sb_list_item_t *pos;
+  float          res;
+
   SB_LIST_FOR_EACH(pos, &opt->values)
   {
     val = SB_LIST_ENTRY(pos, value_t, listitem);
@@ -271,17 +312,24 @@ float sb_get_value_float(char *name)
 }
 
 
-char *sb_get_value_string(char *name)
+float sb_get_value_float(const char *name)
 {
   option_t       *opt;
-  value_t        *val;
-  sb_list_item_t *pos;
 
   opt = find_option(&options, name);
   if (opt == NULL)
-    return NULL;
+    return 0;
 
-  SB_LIST_FOR_EACH(pos, &opt->values)
+  return sb_opt_to_float(opt);
+}
+
+
+char *sb_opt_to_string(option_t *opt)
+{
+  value_t        *val;
+  sb_list_item_t *pos;
+
+  SB_LIST_ONCE(pos, &opt->values)
   {
     val = SB_LIST_ENTRY(pos, value_t, listitem);
     return val->data;
@@ -291,7 +339,7 @@ char *sb_get_value_string(char *name)
 }
 
 
-sb_list_t *sb_get_value_list(char *name)
+char *sb_get_value_string(const char *name)
 {
   option_t       *opt;
 
@@ -299,10 +347,28 @@ sb_list_t *sb_get_value_list(char *name)
   if (opt == NULL)
     return NULL;
 
+  return sb_opt_to_string(opt);
+}
+
+
+sb_list_t *sb_opt_to_list(option_t *opt)
+{
   return &opt->values;
 }
 
 
+sb_list_t *sb_get_value_list(const char *name)
+{
+  option_t       *opt;
+
+  opt = find_option(&options, name);
+  if (opt == NULL)
+    return NULL;
+
+  return sb_opt_to_list(opt);
+}
+
+
 char *sb_print_value_size(char *buf, unsigned int buflen, double value)
 {
   unsigned int i;
@@ -425,7 +491,7 @@ int remove_option(sb_list_t * options, char * optname)
 } 
 
 
-value_t *add_value(sb_list_t *values, char *data)
+value_t *add_value(sb_list_t *values, const char *data)
 {
   value_t *newval;
   
@@ -446,7 +512,7 @@ value_t *add_value(sb_list_t *values, char *data)
 }
 
 
-value_t *find_value(sb_list_t *values, char *data)
+value_t *find_value(sb_list_t *values, const char *data)
 {
   sb_list_item_t *pos;
   value_t        *value; 
@@ -465,7 +531,7 @@ value_t *find_value(sb_list_t *values, char *data)
 }
 
 
-option_t *add_option(sb_list_t *options, char *name)
+option_t *add_option(sb_list_t *options, const char *name)
 {
   option_t *option;
   
@@ -479,6 +545,7 @@ option_t *add_option(sb_list_t *options, char *name)
     return NULL;
 
   option->name = strdup(name);
+  convert_dashes(option->name);
   
   SB_LIST_ADD_TAIL(&option->listitem, options);
 
@@ -486,7 +553,33 @@ option_t *add_option(sb_list_t *options, char *name)
 }
 
 
-option_t *find_option(sb_list_t *options, char *name)
+void convert_dashes(char *s)
+{
+  while (*s != '\0')
+  {
+    if (*s == '-')
+      *s = '_';
+    s++;
+  }
+}
+
+
+int opt_name_cmp(const char *s1, const char *s2)
+{
+  for (/* empty */; *s1 != '\0'; s1++, s2++)
+  {
+    if (*s1 == *s2)
+      continue;
+
+    if ((*s1 != '-' && *s1 != '_') || (*s2 != '-' && *s2 != '_'))
+      break;
+  }
+
+  return *s1 - *s2;
+}
+
+
+option_t *find_option(sb_list_t *options, const char *name)
 {
   sb_list_item_t *pos;
   option_t       *opt;    
@@ -497,7 +590,7 @@ option_t *find_option(sb_list_t *options, char *name)
   SB_LIST_FOR_EACH(pos, options)
   {
     opt = SB_LIST_ENTRY(pos, option_t, listitem);
-    if (!strcmp(opt->name, name))
+    if (!opt_name_cmp(opt->name, name))
       return opt;
   }
   
@@ -505,6 +598,23 @@ option_t *find_option(sb_list_t *options, char *name)
 }
 
 
+sb_list_item_t *sb_options_enum_start(void)
+{
+  return SB_LIST_ENUM_START(&options);
+}
+
+sb_list_item_t *sb_options_enum_next(sb_list_item_t *pos, option_t **opt)
+{
+  pos = SB_LIST_ENUM_NEXT(pos, &options);
+  if (pos == NULL)
+    return NULL;
+
+  *opt = SB_LIST_ENTRY(pos, option_t, listitem);
+
+  return pos;
+}
+
+
 sb_list_t *read_config(FILE *fp, sb_list_t *options)
 {
   char        buf[MAXSTRLEN];
@@ -606,4 +716,3 @@ int write_config(FILE *fp, sb_list_t *options)
     
   return 0;
 }
-
